/*
Copyright (C) 2006-2007 the DSQuake project.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef _LINKSET_H_
#define _LINKSET_H_

#include "LinkSetBase.h"
#include "QP_Network.h"
#include "EndPoint_UDP.h"
#include "ProxyBase.h"

#include <stdio.h>
#include <unistd.h>

struct sockaddr;

class ProxyBase;

template<class RemoteLinkType>
class LinkSet : public LinkSetBase
{
	private:
		int m_Socket;
		EndPoint*m_LocalEndPoint;
		RemoteLinkType**m_RemoteLinks;
		int m_NumRemoteLinks;
		const int m_MaxRemoteLinks;

		char*m_TempDataBuffer;
		const int m_MaxTempDataBuffer;

		char m_Name[QP_MAX_NAME];
		int  m_LinkNameCounter;
		
		ProxyBase*m_ParentProxy;

		virtual void DoReceiveData();
		int FindLinkIndexForRemoteAddress(const EndPoint*in_RemoteEndPoint);
		virtual void DoSendData();
	protected:
		virtual void HandleDataFromAddress(const EndPoint*in_RemoteEndPoint, char*in_Buf, const int in_NumBuf, bool in_bNoRecurse=false);
		virtual void HandleDataFromUnknownAddress(const EndPoint*in_RemoteEndPoint, char*in_Buf, const int in_NumBuf);
	public:
		LinkSet(const char*in_Name, ProxyBase*in_ParentProxy, const EndPoint*in_LocalEndPoint, const int in_MaxLinks = 32);
		virtual int AddRemoteLinkTo(const EndPoint*in_RemoteEndPoint);
		virtual RemoteLink*GetRemoteLinkByNum(int in_Num) const;
		virtual bool HasLocalEndPoint(const EndPoint*in_LocalEndPoint) const;
		bool IsValid() const;
		int GetSocket() const;
		const char*GetName() const;
		virtual void GetText(char*out_strText, const int in_BufLen) const;
		virtual const char*GetTypeName() const;
		virtual ~LinkSet();
		virtual void Work();
		void ForwardData( char*in_Buf, int in_Num );
		template<class RemoteLinkType2>
		void ForwardDataTo( LinkSet<RemoteLinkType2>*in_FwdLinkSet );
		void SetUpForwarding( const EndPoint*in_FromRemoteEndPoint, unsigned int in_LANID, unsigned int in_LinkID);
		ProxyBase*GetParentProxy() const;
};

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::DoReceiveData()
{
	int NumReceived = 0;
	sockaddr_in SourceAddress;
	memset(&SourceAddress, 0, sizeof(sockaddr_in)); 
	int AddrLen = sizeof(sockaddr_in);
	EndPoint_UDP RemoteEndPoint;
	do
	{
		NumReceived = recvfrom(m_Socket, m_TempDataBuffer, m_MaxTempDataBuffer, 0, (sockaddr*)&SourceAddress, &AddrLen);
		if( AddrLen == sizeof(sockaddr_in) )
		{
			RemoteEndPoint.SetAddress((sockaddr_in*)&SourceAddress);
			HandleDataFromAddress(&RemoteEndPoint, m_TempDataBuffer, NumReceived);
		}
	} while(NumReceived>0);
}

template<class RemoteLinkType>
int LinkSet<RemoteLinkType>::FindLinkIndexForRemoteAddress(const EndPoint*in_RemoteEndPoint)
{
	for(int i=0; i<m_NumRemoteLinks; i++)
		if(m_RemoteLinks[i] != NULL && m_RemoteLinks[i]->HasRemoteAddress(in_RemoteEndPoint))
			return i;
	return -1;
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::DoSendData()
{
	for(int i=0; i<m_NumRemoteLinks; i++)
		if(m_RemoteLinks[i] != NULL)
			m_RemoteLinks[i]->DoSendDataWith(m_Socket);
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::HandleDataFromAddress(const EndPoint*in_RemoteEndPoint, char*in_Buf, const int in_NumBuf, bool in_bNoRecurse)
{
	if(in_NumBuf > 0)
	{
		int iLink = FindLinkIndexForRemoteAddress(in_RemoteEndPoint);
		if( iLink >= 0 )
			m_RemoteLinks[iLink]->ReceiveData(in_Buf, in_NumBuf);
		else if(!in_bNoRecurse)
			HandleDataFromUnknownAddress(in_RemoteEndPoint, in_Buf, in_NumBuf);
	}
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::HandleDataFromUnknownAddress(const EndPoint*in_RemoteEndPoint, char*in_Buf, const int in_NumBuf)
{
	char AddrText[32];
	in_RemoteEndPoint->GetText(AddrText, 32);
	printf("%s received %d bytes from unknown party : %s. Ignoring.\n", GetName(), in_NumBuf, AddrText);
}

template<class RemoteLinkType>
LinkSet<RemoteLinkType>::LinkSet(const char*in_Name, ProxyBase*in_ParentProxy, const EndPoint*in_LocalEndPoint, const int in_MaxLinks)
: m_Socket(0), m_LocalEndPoint(in_LocalEndPoint->Clone()),
  m_RemoteLinks(NULL), m_NumRemoteLinks(0), m_MaxRemoteLinks(in_MaxLinks), 
  m_MaxTempDataBuffer(65536), m_LinkNameCounter(0),
  m_ParentProxy(in_ParentProxy)
{
	strncpy(m_Name, in_Name, QP_MAX_NAME);
	m_RemoteLinks = new RemoteLinkType*[m_MaxRemoteLinks];

	m_Socket = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(m_Socket < 0)
	{
		printf("%s [%s]: Error creating UDP socket.\n", GetName(), GetTypeName());
		return;
	}
	int One = 1;
	ioctlsocket(m_Socket,FIONBIO,(u_long*)&One);
	int res = m_LocalEndPoint->BindSocketLocal(this->GetSocket());
	if(res != 0)
	{
		char AddressText[32];
		m_LocalEndPoint->GetText(AddressText, 32);
		printf("%s [%s]: Error binding socket to local address %s.\n", GetName(), GetTypeName(), AddressText);
		return;
	}
	m_TempDataBuffer = new char[m_MaxTempDataBuffer]; 
}

template<class RemoteLinkType>
int LinkSet<RemoteLinkType>::AddRemoteLinkTo(const EndPoint*in_RemoteEndPoint)
{
	int nLink = FindLinkIndexForRemoteAddress(in_RemoteEndPoint);

	if( nLink >= 0 )
		return nLink;

	if( m_NumRemoteLinks<m_MaxRemoteLinks )
	{
		char LinkName[QP_MAX_NAME];
		snprintf(LinkName, QP_MAX_NAME, "%s.L%04d", GetName(), m_LinkNameCounter++);
		m_RemoteLinks[m_NumRemoteLinks] = new RemoteLinkType(LinkName, this, m_ParentProxy->AssignNewLinkID(), m_LocalEndPoint, in_RemoteEndPoint->Clone());
		m_NumRemoteLinks++;
		return m_NumRemoteLinks-1;
	}
	return -1;
}

template<class RemoteLinkType>
RemoteLink*LinkSet<RemoteLinkType>::GetRemoteLinkByNum(int in_Num) const
{
	if(in_Num >= 0 && in_Num < m_MaxRemoteLinks)
		return m_RemoteLinks[in_Num];
	return NULL;
}

template<class RemoteLinkType>
bool LinkSet<RemoteLinkType>::IsValid() const
{
	for(int i=0; i<m_NumRemoteLinks; i++)
		if(m_RemoteLinks[i] != NULL && !m_RemoteLinks[i]->IsValid())
			return false;
	if(m_LocalEndPoint == NULL || !m_LocalEndPoint->IsValid() )
		return false;
	return (m_Socket>0);
}

template<class RemoteLinkType>
bool LinkSet<RemoteLinkType>::HasLocalEndPoint(const EndPoint*in_LocalEndPoint) const
{
	return IsValid() && m_LocalEndPoint->HasSameAddress(in_LocalEndPoint); 
}

template<class RemoteLinkType>
int LinkSet<RemoteLinkType>::GetSocket() const
{
	if(!IsValid())
		return -1;
	return m_Socket;
}

template<class RemoteLinkType>
const char*LinkSet<RemoteLinkType>::GetName() const
{
	return m_Name;
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::GetText(char*out_strText, const int in_BufLen) const  
{
	if(!IsValid())
	{
		snprintf(out_strText, in_BufLen, "%s [%s]: Invalid state.", GetName(), GetTypeName());
		return;
	}
	char strAddress[32];
	m_LocalEndPoint->GetText(strAddress, 32);
	snprintf(out_strText, in_BufLen, "%s [%s] from %s with %d/%d links:\n", GetName(), GetTypeName(), strAddress, m_NumRemoteLinks, m_MaxRemoteLinks);
	char TextBuf[2048];
	const char*LinkText;
	for(int i=0; i<m_NumRemoteLinks; i++)
	{
		LinkText = ( m_RemoteLinks[i] != NULL ? m_RemoteLinks[i]->GetText(TextBuf, 2048), TextBuf : "NULL" );
		snprintf(out_strText, in_BufLen, "%s\t%d:[%s]\n", out_strText, i, LinkText);
	}
}

template<class RemoteLinkType>
const char*LinkSet<RemoteLinkType>::GetTypeName() const
{
	return "LinkSet";
}

template<class RemoteLinkType>
LinkSet<RemoteLinkType>::~LinkSet()
{
	if( m_RemoteLinks != NULL )
	{
		for(int i=0; i<m_NumRemoteLinks; i++)
			if( m_RemoteLinks[i] != NULL )
				delete m_RemoteLinks[i];
		delete[] m_RemoteLinks;
	}
	if(m_LocalEndPoint!=NULL)
		delete m_LocalEndPoint;
	if(m_Socket>=0)
		close(m_Socket);
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::Work()
{
	if(!IsValid())
		return;
	DoReceiveData();
	for(int i=0; i<m_NumRemoteLinks; i++)
		if( m_RemoteLinks[i] != NULL )
			m_RemoteLinks[i]->Work();
	DoSendData();
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::ForwardData( char*in_Buf, int in_Num )
{
	if(in_Num>0)
		for(int i=0; i<m_NumRemoteLinks; i++)
			m_RemoteLinks[i]->SendForwardedData(in_Buf, in_Num);
}

template<class RemoteLinkType>
template<class RemoteLinkType2>
void LinkSet<RemoteLinkType>::ForwardDataTo( LinkSet<RemoteLinkType2>*in_FwdLinkSet )
{
	for(int i=0; i<m_NumRemoteLinks; i++)
	{
		char*Buf = NULL;
		int  Num = 0;
		m_RemoteLinks[i]->GetForwardData(Buf, Num);
		in_FwdLinkSet->ForwardData(Buf, Num);
		m_RemoteLinks[i]->ClearForwardData();
	}
}

template<class RemoteLinkType>
void LinkSet<RemoteLinkType>::SetUpForwarding( const EndPoint*in_FromRemoteEndPoint, unsigned int in_LANID, unsigned int in_LinkID)
{
	int nLink = AddRemoteLinkTo(in_FromRemoteEndPoint);
	if( nLink >= 0 )
		m_RemoteLinks[nLink]->SetForwardingDestination(in_LANID, in_LinkID);		
}

template<class RemoteLinkType>
ProxyBase*LinkSet<RemoteLinkType>::GetParentProxy() const
{
	return m_ParentProxy;
}

#endif // _LINKSET_H_
