/*
Copyright (C) 2006-2007 the DSQuake project.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "RemoteLink.h"
#include "EndPoint.h"
#include "LinkSetBase.h"

#include <winsock2.h>
#include <stdio.h>

RemoteLink::RemoteLink(const char*in_Name, LinkSetBase*in_ParentLinkSet, unsigned int in_LinkID, const EndPoint*in_LocalEndPoint, EndPoint*in_RemoteEndPoint)
:m_bValid(false),
 m_NumSendBuffer(0), m_MaxSendBuffer(65536), m_SendBuffer(NULL),
 m_NumRecvBuffer(0), m_MaxRecvBuffer(65536), m_RecvBuffer(NULL),
 m_LocalEndPoint(in_LocalEndPoint), m_RemoteEndPoint(in_RemoteEndPoint), m_ParentLinkSet(in_ParentLinkSet),
 m_LinkID(in_LinkID)
 {
	strncpy(m_Name, in_Name, QP_MAX_NAME);
	if(m_LocalEndPoint == NULL || !m_LocalEndPoint->IsValid() || m_RemoteEndPoint == NULL || !m_RemoteEndPoint->IsValid())
	{
		printf("Error creating remote link %s: invalid end point.\n", GetName());
		return;
	}
	m_SendBuffer = new char[m_MaxSendBuffer];
	m_RecvBuffer = new char[m_MaxRecvBuffer];
	m_bValid=true;
}

bool RemoteLink::IsValid() const
{
	return m_bValid;
}

const char*RemoteLink::GetName() const
{
	return m_Name;
}

int RemoteLink::GetPacketLength( const char*in_Data, int in_NumData ) 
{
	return in_NumData;
}

void RemoteLink::GetText(char*out_strText, const int in_BufLen) const  
{
	if(!IsValid())
	{
		snprintf(out_strText, in_BufLen, "%s [%s]: Invalid state.", GetName(), GetTypeName());
		return;
	}
	char strAddress1[32],strAddress2[32];
	m_LocalEndPoint->GetText(strAddress1, 32);
	m_RemoteEndPoint->GetText(strAddress2, 32);
	snprintf(out_strText, in_BufLen, "%s [%s]: %s -> %s. ID=%d.", GetName(), GetTypeName(), strAddress1, strAddress2, m_LinkID);
}

bool RemoteLink::HasRemoteAddress(const EndPoint*in_RemoteEndPoint) const
{
	return IsValid() && in_RemoteEndPoint->IsValid() && m_RemoteEndPoint->HasSameAddress(in_RemoteEndPoint);
}

void RemoteLink::DoSendDataWith(int in_Socket)
{
	int nPacket = 0;
	while( (nPacket = GetPacketLength( m_SendBuffer, m_NumSendBuffer )) > 0 && nPacket <= m_NumSendBuffer )
	{
		int DataSent = m_RemoteEndPoint->SendDataWith(in_Socket, m_SendBuffer, nPacket );
		if( DataSent > 0 )
		{
			if( DataSent < m_NumSendBuffer )
				memmove( m_SendBuffer, m_SendBuffer+DataSent, m_NumSendBuffer-DataSent );
			m_NumSendBuffer-=DataSent;
		}
	}
}

void RemoteLink::ReceiveData(const char*in_RecvBuf, const int in_NumRecvBuf)
{
	int NewBytes = in_NumRecvBuf;
	if( in_NumRecvBuf > m_MaxRecvBuffer - m_NumRecvBuffer )
		NewBytes = m_MaxRecvBuffer - m_NumRecvBuffer;
	if( NewBytes > 0 )
		memcpy( m_RecvBuffer + m_NumRecvBuffer, in_RecvBuf, NewBytes );
	m_NumRecvBuffer+=NewBytes;  
	//printf("%s: Received %d bytes.\n", GetName(), NewBytes); 
}

void RemoteLink::SendData(const char*in_SendBuf, const int in_NumSendBuf)
{
	int NewBytes = in_NumSendBuf;
	if( in_NumSendBuf > m_MaxSendBuffer - m_NumSendBuffer )
		NewBytes = m_MaxSendBuffer - m_NumSendBuffer;
	if( NewBytes > 0 )
		memcpy( m_SendBuffer + m_NumSendBuffer, in_SendBuf, NewBytes );
	m_NumSendBuffer+=NewBytes;
}

const char*RemoteLink::GetTypeName() const
{
	return "RemoteLink";
}

RemoteLink::~RemoteLink()
{
	// LocalEndPoint destroyed by owner LinkSet
	if(m_RemoteEndPoint!=NULL)
		delete m_RemoteEndPoint;
	if(m_SendBuffer!=NULL)
		delete[]m_SendBuffer;
	if(m_RecvBuffer!=NULL)
		delete[]m_RecvBuffer;
}

ProxyBase*RemoteLink::GetParentProxy() const
{
	return m_ParentLinkSet?m_ParentLinkSet->GetParentProxy():NULL;
}

unsigned int RemoteLink::GetLinkID() const
{
	return m_LinkID;
}

void RemoteLink::SendForwardedData(unsigned int in_SrcLANID, unsigned int in_SrcLinkID, unsigned int in_DstLANID, unsigned int in_DstLinkID, int in_ReplaceLocalPortAtOffset, const char*in_FwdBuf, int in_NumFwdBuf)
{
	//printf("%s: Sending a %dB forwarded message from %d/%d.\n", GetName(), in_NumFwdBuf, in_SrcLANID, in_SrcLinkID);
	SendData(in_FwdBuf, in_NumFwdBuf);
}

const EndPoint*RemoteLink::GetLocalEndPoint() const
{
	return m_LocalEndPoint;
}

void RemoteLink::Work()
{
}
