/*
Copyright (C) 2006-2007 the DSQuake project.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "TunnelLink.h"
#include "ProxyBase.h"

#include <stdio.h>
#include <mem.h>

void TunnelLink::SendTunnelMessage( unsigned char in_Code, const char*in_Data, int in_NumData )
{
	char Buffer[QP_MSG_HEADER_SIZE];
	*((unsigned short*)(Buffer+QP_MSG_SIZE_OFFSET)) = in_NumData + QP_MSG_HEADER_SIZE;
	*((unsigned char *)(Buffer+QP_MSG_CODE_OFFSET)) = in_Code;
	SendData( Buffer, QP_MSG_HEADER_SIZE );
	if( in_NumData > 0 && in_Data != NULL )
		SendData( in_Data, in_NumData );
}

void TunnelLink::ProcessReceivedData()
{
	while( m_NumRecvBuffer >= QP_MSG_HEADER_SIZE ) 
	{
		int MsgSize = *((unsigned short*)(m_RecvBuffer+QP_MSG_SIZE_OFFSET));
		if ( MsgSize > m_NumRecvBuffer ) 
		{
			//printf("%s: Warning : got partial message of %d/%d bytes! Discarding.\n", GetName(), m_NumRecvBuffer, MsgSize);
			//m_NumRecvBuffer = 0;
			return;
		}
		else
		{
			TunnelMessageCode MsgCode = (TunnelMessageCode)*((unsigned char *)(m_RecvBuffer+QP_MSG_CODE_OFFSET));
			switch( MsgCode )
			{
				case TMC_KeepAlive:
				{
					//printf("%s: Received KeepAlive !\n", GetName());
					int CurTime = GetTimeMSecs();
					m_LastReceivedKeepAliveTimeStamp = CurTime;
					if( MsgSize > QP_MSG_HEADER_SIZE )
						printf("%s: Warning : got %d bytes of data with KeepAlive message! Discarding. (%d bytes in received queue)\n", GetName(), MsgSize, m_NumRecvBuffer);
				}
					break;
				case TMC_Forward:
				{
					unsigned int SrcLANID    = m_LinkID;
					unsigned int SrcLinkID   = *((unsigned int*)(m_RecvBuffer+QP_MSG_HEADER_SIZE+QP_MSG_FORWARD_SRC_LINKID_OFFSET  ));
					//unsigned int DstLANID    = *((unsigned int*)(m_RecvBuffer+QP_MSG_HEADER_SIZE+QP_MSG_FORWARD_DST_LANID_OFFSET   ));
					unsigned int DstLANID    = QP_LAN_ID_SELF; 
					unsigned int DstLinkID   = *((unsigned int*)(m_RecvBuffer+QP_MSG_HEADER_SIZE+QP_MSG_FORWARD_DST_LINKID_OFFSET  ));
					short        PortReplace = *((short       *)(m_RecvBuffer+QP_MSG_HEADER_SIZE+QP_MSG_FORWARD_PORT_REPLACE_OFFSET));
					char*FwdMsgData = m_RecvBuffer+QP_MSG_HEADER_SIZE+QP_MSG_FORWARD_HEADER_SIZE;
					int  FwdMsgSize = MsgSize-QP_MSG_HEADER_SIZE-QP_MSG_FORWARD_HEADER_SIZE;
					RemoteLink*NextLink = GetParentProxy()->GetNextLinkInRoute( DstLANID, DstLinkID );
					GetParentProxy()->AddRoutingEntry( SrcLANID, SrcLinkID, this );
					if( NextLink != NULL )
					{
						//printf("%s: Received Forward message of %dB coming from %d/%d and destined to %d/%d (%s)!\n", GetName(), FwdMsgSize, SrcLANID, SrcLinkID, DstLANID, DstLinkID, NextLink->GetName());
						NextLink->SendForwardedData( SrcLANID, SrcLinkID, m_ForwardLANID, m_ForwardLinkID, PortReplace, FwdMsgData, FwdMsgSize );
					}
					else
					{
						printf("%s: WARNING: Received Forward message of %dB coming from %d/%d and destined to %d/%d -> COULDN'T FIND ROUTE TO DESTINATION!\n", GetName(), FwdMsgSize, SrcLANID, SrcLinkID, DstLANID, DstLinkID);
					}
				}
					break;
				default:
					printf("%s: Warning : got unknown message code %d, in %d-byte message! (%d bytes in received queue)\n", GetName(), MsgCode, MsgSize, m_NumRecvBuffer);
			}
			if( m_NumRecvBuffer > MsgSize )
			{
				memmove( m_RecvBuffer, m_RecvBuffer + MsgSize, m_NumRecvBuffer - MsgSize );
			}
			m_NumRecvBuffer -= MsgSize;
		}
	}
	if( m_NumRecvBuffer > 0 )
	{
		printf("%s: Warning : got %d unmatched bytes! Discarding.\n", GetName(), m_NumRecvBuffer);
		m_NumRecvBuffer = 0; 
	}
}

TunnelLink::TunnelLink(const char*in_Name, LinkSetBase*in_ParentLinkSet, unsigned int in_LinkID, const EndPoint*in_LocalEndPoint,EndPoint*in_RemoteEndPoint)
:ForwardedLink(in_Name, in_ParentLinkSet, in_LinkID, in_LocalEndPoint, in_RemoteEndPoint),
 m_LastKeepAliveTimeStamp(0), m_KeepAliveDelay(3000), m_LastReceivedKeepAliveTimeStamp(0), m_bIsAlive(false)
{
}

const char*TunnelLink::GetTypeName() const
{
	return "TunnelLink";
}

void TunnelLink::Work()
{
	int CurTime = GetTimeMSecs();
	ProcessReceivedData();

	bool bWasAlive = m_bIsAlive;
	m_bIsAlive = m_LastReceivedKeepAliveTimeStamp > 0 && (CurTime - m_LastReceivedKeepAliveTimeStamp) < 2.1*m_KeepAliveDelay;
	if( bWasAlive != m_bIsAlive )
		printf("%s: Tunnel link %s.\n", GetName(), m_bIsAlive?"connected":"disconnected");

	if( (CurTime - m_LastKeepAliveTimeStamp) > m_KeepAliveDelay || (m_bIsAlive && !bWasAlive) )
	{
		//printf("%s: sending KeepAlive.\n", GetName());
		SendTunnelMessage(TMC_KeepAlive, NULL, 0);
		m_LastKeepAliveTimeStamp = CurTime;
	}

	RemoteLink::Work();
}
/*
void TunnelLink::SendForwardedData(char*in_FwdData, int in_NumFwdData)
{
	SendTunnelMessage(TMC_Forward, in_FwdData, in_NumFwdData);
}*/

void TunnelLink::SendForwardedData(unsigned int in_SrcLANID, unsigned int in_SrcLinkID, unsigned int in_DstLANID, unsigned int in_DstLinkID, int in_ReplaceLocalPortAtOffset, const char*in_FwdBuf, int in_NumFwdBuf)
{
	char*Fwd_Msg_Header = new char[QP_MSG_FORWARD_HEADER_SIZE + in_NumFwdBuf];
	*((unsigned int*)(Fwd_Msg_Header+QP_MSG_FORWARD_SRC_LINKID_OFFSET  )) = in_SrcLinkID;
	//*((unsigned int*)(Fwd_Msg_Header+QP_MSG_FORWARD_DST_LANID_OFFSET   )) = in_DstLANID;
	*((unsigned int*)(Fwd_Msg_Header+QP_MSG_FORWARD_DST_LINKID_OFFSET  )) = in_DstLinkID;
	*((short       *)(Fwd_Msg_Header+QP_MSG_FORWARD_PORT_REPLACE_OFFSET)) = (short)in_ReplaceLocalPortAtOffset;
	memcpy(Fwd_Msg_Header + QP_MSG_FORWARD_HEADER_SIZE, in_FwdBuf, in_NumFwdBuf);
	//printf("%s: Sending Forward Tunnel message of %dB coming from %d/%d and destined to %d/%d!\n", GetName(), in_NumFwdBuf, in_SrcLANID, in_SrcLinkID, in_DstLANID, in_DstLinkID);
	SendTunnelMessage(TMC_Forward, Fwd_Msg_Header, QP_MSG_FORWARD_HEADER_SIZE + in_NumFwdBuf);
}
